class EmptyTreeError(Exception):
    def __init__(self):
        super().__init__('empty tree')


class InvalidEngineError(Exception):
    def __init__(self, message):
        super().__init__(f"invalid engine: {message}")


class InvalidJSONError(Exception):
    pass


class InvalidMerkleProofError(Exception):
    def __init__(self, message):
        super().__init__(f"invalid proof: {message}")


class TreeNotBuildError(Exception):
    def __init__(self):
        super().__init__('tree not built')


class UnableToBuildPathError(Exception):
    def __init__(self, message) -> None:
        super().__init__(f"unable to build path, found: {message}")



import base64
from pyutls.hex import from_hex, to_hex
import re

from merklepy.exception import InvalidMerkleProofError
from merklepy.hash import SHA_256, Hashes
from merklepy.path import Path


class Proof(object):
    """
    The proof consists of the trail of intermediate hashes suffixed with the path, the name
    of the hashing engine used and the size of the Merkle tree at the date of the proof.

    Parameters
    ----------
    trail : Hashes
        The intermediate hashes bottom up (not including the final root hash)
    path : Path
        The path from root to leaf
    size : int
        The number of leaves in the Merkle tree at issuance of the proof
    engine : str, optional
        The name of the hashing function (the default is `'sha-256'`)
    """

    def __init__(self, trail: Hashes, path: Path, size: int, engine: str = SHA_256):
        self.trail = trail
        self.path = path
        self.size = size
        self.engine = engine

    def to_string(self) -> str:
        """
        Returns
        -------
        str
            The base64-encoded dot-separated concatenation of the hexadecimal hashes, the path, the engine and the size of the tree, eg.

            Base64('&lt;hash1&gt;&lt;hash2&gt;.11.sha-256.4')
        """
        trail_hex = ''.join(map(to_hex, self.trail))
        readable_proof = '.'.join(
            [trail_hex, self.path, self.engine, str(self.size)])
        return base64.b64encode(bytes(readable_proof, 'utf-8')).decode('utf-8')


def proof_from(string: str) -> Proof:
    """Build a `Proof` from the passed string, provided it's an actual stringified proof"""
    try:
        decoded = str(base64.b64decode(string), 'utf-8')
    except Exception as error:
        raise InvalidMerkleProofError(string) from error
    parts = decoded.split('.')
    if len(parts) != 4:
        raise InvalidMerkleProofError(string)
    path = parts[1]
    engine = parts[2]
    size = int(parts[3])
    if size == 0:
        raise InvalidMerkleProofError(string)
    if engine == SHA_256:
        hashes = list(filter(lambda item: len(item) == 32,
                             map(from_hex, re.split('(.{64})', parts[0]))))
        return Proof(hashes, path, size, engine)
    else:
        raise InvalidMerkleProofError(string)



import json

from merklepy.hash import SHA_256


class MerkleTreeOptions(object):
    """
    Define the `MerkleTree` options

    Parameters
    ----------
    doubleHash : bool, optional
        Set to `True` to hash nodes twice (the default is `False`).
    engine : str, optional
        The name of the hashing function (the default is `'sha-256'`).
    sort : bool, optional
        Set to `True` to lexicographically sort leaves (the default is `False`).
    """

    def __init__(self, doubleHash: bool = False, engine: str = SHA_256, sort: bool = False):
        self.doubleHash = doubleHash
        self.engine = engine
        self.sort = sort

    def to_json(self) -> str:
        """Returns the JSON-stringified representation of the current `MerkleTreeOptions`"""
        return json.dumps(self.__dict__, separators=(',', ':'))



import math

from merklepy.exception import UnableToBuildPathError

Path = str

LEFT = '1'
RIGHT = '0'


def _half_bucket(using: int) -> float:
    return math.pow(2, math.ceil(math.log(using) / math.log(2))) / 2


def build_path(index: int, size: int, depth: int) -> Path:
    """
    Build the left-right path from the root to the leaf at the passed index.

    Parameters
    ----------
    index : int
        The leaf index.
    size : int
        The size of the Merkle tree.
    depth : int
        The depth of the Merkle tree.

    Returns
    -------
    Path
        The path code.

    Raises
    ------
    UnableToBuildPathError
        If the path length is different from the tree depth.
    """
    path = ''
    initial_depth = depth
    while size > 0 and depth > 0:
        half = _half_bucket(size)
        if index < half:
            path += LEFT
            size = max(half, 1)
        else:
            path += RIGHT
            index -= half
            if size - half * 2 == 0:
                size = max(half, 1)
            else:
                size = size - half
        depth -= 1
    if len(path) != initial_depth:
        raise UnableToBuildPathError(path)
    return path
